class EnumMember:

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    def __new__(cls, enum_name, name, value):
        instance = object.__new__(cls)
        instance._name = name
        instance._value = value
        instance._enum_name = enum_name
        return instance

    def __repr__(self):
        return f'{self._enum_name}{{{self._name}}}'

    def __eq__(self, other):
        return self is other or self._value == other

    def __ne__(self, other):
        return not self.__eq__(other)



import asyncio
import logging
from ..client import Client
from ..client.abc.events import Events


class Emitter(Events):

    _ev_handlers = dict()

    def __init__(self, client: Client, emitter: str = '', scope: str = None):
        """Initializes an emitter.

        Args:
            client (thingsdb.client.Client):
                ThingsDB Client instance.
            emitter (str):
                Code which should point to the `thing` to watch for events.
                Defaults to an empty string which is the collection.
                Examples are:
                   - ''
                   - '.emitter'
                   - '#123'
            scope (str):
                Collection scope. Defaults to the scope of the client.
        """
        super().__init__()
        self._event_id = 0
        self._client = client
        self._thing_id = None
        self._scope = scope or client.get_scope()
        if emitter == '.':
            emitter = ''
        if emitter:
            emitter = f'{{{emitter}}}'
        self._code = \
            f'{emitter}.watch(); {emitter}.id();'
        client.add_event_handler(self)
        asyncio.ensure_future(self._watch())

    def __init_subclass__(cls):
        cls._ev_handlers = {}

        for key, val in cls.__dict__.items():
            if not key.startswith('__') and \
                    callable(val) and hasattr(val, '_ev'):
                cls._ev_handlers[val._ev] = val

    async def _watch(self):
        self._thing_id = await self._client.query(
            self._code,
            scope=self._scope)

    def on_reconnect(self):
        asyncio.ensure_future(self._watch())

    def on_node_status(self, _status):
        pass

    def on_warning(self, warn):
        logging.warning(f'{warn["warn_msg"]} ({warn["warn_code"]})')

    def on_watch_init(self, _data):
        pass

    def on_event(self, ev, *args):
        cls = self.__class__
        fun = cls._ev_handlers.get(ev)
        if fun is None:
            logging.debug(f'no event handler for {ev} on {cls.__name__}')
            return
        fun(self, *args)

    def on_watch_update(self, data):
        thing_id = data['#']
        if thing_id != self._thing_id:
            return

        event_id, jobs = data['event'], data['jobs']

        if self._event_id > event_id:
            logging.warning(
                f'ignore event because the current event `{self._event_id}` '
                f'is greather than the received event `{event_id}`')
            return
        self._event_id = event_id

        for job_dict in jobs:
            for name, job in job_dict.items():
                if name == 'event':
                    self.on_event(*job)

    def on_watch_delete(self, data):
        thing_id = data['#']
        if thing_id == self._thing_id:
            logging.debug(f'emitter with id {thing_id} is removed')
            self._client.remove_event_handler(self)

    def on_watch_stop(self, data):
        thing_id = data['#']
        if thing_id == self._thing_id:
            logging.debug(f'emitter with id {thing_id} is stopped')
            self._client.remove_event_handler(self)



from .enummember import EnumMember

_enums_lookup = {}  # enums lookup by name


class _GetAttr(type):
    def __getitem__(cls, name):
        for k, v in cls.__dict__.items():
            if k == name:
                return v
        raise KeyError(f'no member with name `{k}`')


class Enum(metaclass=_GetAttr):

    _visited = 0  # For build, 0=not visited, 1=new_type, 2=set_type, 3=build

    def __new__(cls, *args):
        if len(args) != 1:
            return super().__new__(cls)

        value = args[0]
        for v in cls.__dict__.values():
            if isinstance(v, EnumMember) and v._value == value:
                return v
        raise ValueError(f'no member with value `{value}`')

    def __init_subclass__(cls, **kwargs):
        if issubclass(cls, EnumMember):
            return

        cls._name = getattr(cls, '__NAME__', cls.__name__)
        cls._cname = getattr(cls, '__COLLECTION_NAME__', '')
        cls._id = None
        cls._memberclass = type(f'{cls._name}Member', (EnumMember, cls), {})

        # upgrade attributes to member instances
        for k, v in cls.__dict__.items():
            if k.startswith('_'):
                continue
            setattr(cls, k, cls._memberclass(cls._name, k, v))

        # register for lookup by name
        _enums_lookup[cls._cname + cls._name] = cls

    @staticmethod
    def _update_enum(cname, enums, data, convert):
        name = data['name']
        enum = _enums_lookup.get(name, _enums_lookup.get(cname + name))
        if enum and enum._cname and enum._cname != cname:
            raise TypeError(
                f'Enum type {name} is used in more than one collection; '
                'Add __COLLECTION_NAME__ to your Enum definition to fix '
                'this error')
        cls = EnumMember if enum is None else enum._memberclass

        members = [cls(name, k, convert(v)) for k, v in data['members']]

        if enum is not None:
            enum._cname = cname
            enum._id = data['enum_id']
            for member in members:
                setattr(enum, member.name, member)

        enums[data['enum_id']] = members, enum

    @staticmethod
    def _upd_enum_add(enums, data, convert):
        members, enum = enums[data['enum_id']]
        name = members[0]._enum_name
        cls = EnumMember if enum is None else enum._memberclass

        member = cls(name, data['name'], convert(data['value']))
        members.append(member)

        if enum is not None:
            setattr(enum, member.name, member)

    @staticmethod
    def _upd_enum_del(enums, data):
        members, enum = enums[data['enum_id']]

        if enum is not None:
            member = members[data['index']]
            delattr(enum, member.name)

        try:
            # swap remove the index
            members[data['index']] = members.pop()
        except IndexError:
            pass

    @staticmethod
    def _upd_enum_def(enums, data):
        members, _ = enums[data['enum_id']]
        # swap index
        idx = data['index']
        tmp = members[idx]
        members[idx] = members[0]
        members[0] = tmp

    @staticmethod
    def _upd_enum_mod(enums, data, convert):
        members, _ = enums[data['enum_id']]
        name = members[0]._enum_name
        member = members[data['index']]
        member._value = convert(data['value'])

    @staticmethod
    def _upd_enum_ren(enums, data):
        members, enum = enums[data['enum_id']]
        member = members[data['index']]

        if enum is not None:
            delattr(enum, member.name)
            setattr(enum, data['name'], member)

        member._name = data['name']

    @classmethod
    async def _new_type(cls, client, collection):
        if cls._visited > 0:
            return
        cls._visited += 1
        members = (
            m for m in cls.__dict__.values()
            if isinstance(m, EnumMember))
        query = f'''
            set_enum('{cls._name}', {{
                {', '.join(f'{m.name}: {m.value!r}' for m in members)}
            }});
        '''
        await client.query(query, scope=collection._scope)

    @classmethod
    async def _set_type(cls, client, collection):
        pass

    @classmethod
    def id(cls):
        return cls._id

    @classmethod
    def name(cls):
        return cls._name


"""Decorator for handleing events."""

def event(ev):

    def _event(fun):
        def wrapper(self, *args):
            fun(self, *args)

        wrapper._ev = ev
        return wrapper

    return _event
